from abc import ABC, abstractmethod
from time import time
from traceback import format_exc
from typing import Dict, List, Optional

from schema import Schema, SchemaError


class Operation(ABC):
    def __init__(self, options: Dict = None, entry_phase: str = None, schema: Optional[Schema] = None):
        self.operation_time = time()
        self.options = options if options else {}
        self._entry_phase = entry_phase
        self.schema: Optional[Schema] = schema
        self.current_phase = None
        self.break_phase = None
        self.fail_phase = None
        self.success = False
        self.fail_traceback: Optional[str] = None
        self.fail_message: Optional[str] = None

    def run(self):
        try:
            if self.schema:
                self.schema.validate(self.options)
            self._run_phases()
        except SchemaError as e:
            self.on_exception(exception=e)
        except Exception as e:
            if self.success:
                self.break_phase = self.current_phase
            else:
                self.on_exception(exception=e)

        self.operation_time = time() - self.operation_time
        return self

    def _run_phases(self):
        phases = self.phases()
        if self._entry_phase:
            start_phase = phases.index(self._entry_phase)
            phases = phases[start_phase:]
        for phase in phases:
            self.current_phase = phase
            getattr(self, phase)()
        self.success = True

    @abstractmethod
    def phases(self) -> List[str]:
        raise NotImplementedError

    def break_operation(self, message: Optional[str] = None):
        self.success = True
        raise Exception(message)

    def fail_operation(self, message: Optional[str] = None):
        self.success = False
        raise Exception(message)

    def on_exception(self, exception: Optional[Exception] = None):
        self.fail_phase = self.current_phase
        self.fail_traceback = format_exc()
        self.fail_message = str(exception) if exception else None



from operation import Operation
from schema import Schema


class TestOperation(Operation):

    def phases(self):
        return [
            'phase_one',
            'phase_two',
            'phase_three'
        ]

    def phase_one(self):
        self.options['option_for_second_phase'] = True

    def phase_two(self):
        if 'option_for_second_phase' in self.options:
            return 'an option got passed from first phase, cool!'

    def phase_three(self):
        self.options['reached_third_phase'] = True

    def phase_test_break(self):
        if 'some_option' not in self.options:
            self.break_operation()

    def phase_test_fail(self):
        if 'another_option' not in self.options:
            self.fail_phase()


def run_operation(initial_options=None, phases=None, entry_phase=None, schema=None):
    wo = TestOperation(options=initial_options if initial_options else {}, entry_phase=entry_phase, schema=schema)
    if phases:
        wo.phases = lambda: phases
    return wo.run()


def test_happy_trail():
    wo = run_operation()
    assert wo.success is True
    assert 'option_for_second_phase' in wo.options
    assert 'reached_third_phase' in wo.options


def test_break_operation():
    phases_with_break = ['phase_one', 'phase_two', 'phase_test_break', 'phase_three']
    wo = run_operation(initial_options={'initial_option_1': 1}, phases=phases_with_break)
    assert wo.success is True
    assert 'option_for_second_phase' in wo.options
    assert wo.break_phase is 'phase_test_break'
    assert 'reached_third_phase' not in wo.options


def test_fail_operation():
    phases_with_fail = ['phase_one', 'phase_two', 'phase_test_fail', 'phase_three']
    wo = run_operation(initial_options={'initial_option_1': 1}, phases=phases_with_fail)
    assert wo.success is False
    assert wo.fail_traceback is not None
    assert wo.fail_message is not None
    assert 'option_for_second_phase' in wo.options
    assert wo.fail_phase is 'phase_test_fail'
    assert 'reached_third_phase' not in wo.options


def test_entry_phase():
    wo = run_operation(initial_options={'some_initial_options': 1}, entry_phase='phase_three')
    assert wo.success is True
    assert 'option_for_second_phase' not in wo.options
    assert 'reached_third_phase' in wo.options


class TestSchemaValidation:

    class TestWhenSchemaIsValid:
        @staticmethod
        def test_it_should_run_phase_successfully():
            schema = Schema({'user_id': int, 'logged_in': bool})
            event_payload = {'user_id': 123, 'logged_in': True}
            wo = run_operation(initial_options=event_payload, schema=schema)
            assert wo.success is True

    class TestWhenSchemaIsInvalid:
        @staticmethod
        def test_it_should_fail_operation_immediately():
            schema = Schema({'user_id': int, 'logged_in': bool})
            event_payload = {'user_id': 'not a number', 'logged_in': True}
            wo = run_operation(initial_options=event_payload, schema=schema)
            assert wo.success is False
            assert wo.fail_traceback is not None
            assert 'option_for_second_phase' not in wo.options
